#import "KeyController.h"
#import "MailContents.h"
#import "KeyMethod.h"
#import "KeyWindow.h"
#import "MethodKey.h"
#import "Utils.h"
#import "BrowserTextView.h"
#import "MailTitles.h"
#import "MenuController.h"

@implementation KeyController

- (void) awakeFromNib
{
NSArray *saveKeyArray;

	[[NSNotificationCenter defaultCenter] addObserver: self
		selector: @selector(terminate:)
		name: NSApplicationWillTerminateNotification object:nil];
	[self makeMethodArray];

	keyMethodDict = [[NSMutableDictionary alloc] initWithCapacity: 1];

	if (saveKeyArray = [[NSUserDefaults standardUserDefaults]
							arrayForKey: @"Key Bind"])
	{
		[self loadKeyMethodDict: saveKeyArray];
		//[saveKeyArray autorelease];
				/*array from UserDefaults is already temporal*/
	}
	else
	{
		[self keyMethodInit];
	}
	keyArray = [[keyMethodDict allKeys] mutableCopy];
	[keyArray sortUsingFunction: SortSerial context: keyMethodDict];
	if ([keyArray count])
		[keyTable selectRow: 0 byExtendingSelection: NO];
}


- (void) dealloc
{
	[keyMethodDict release];
	[keyArray release];
	[methodArray release];
	[super dealloc];
}


- (void) terminate: (NSNotification *) notification
{
NSMutableArray *saveKeyArray;
NSEnumerator *enumerator;
NSArray *methodKeyArray;
KeyMethod *keyMethod;
NSArray *tmpArray;

	saveKeyArray = [NSMutableArray arrayWithCapacity: 1];
	enumerator = [keyArray objectEnumerator];
	
	while (methodKeyArray = [enumerator nextObject])
	{
		keyMethod = [keyMethodDict objectForKey: methodKeyArray];
		tmpArray = [NSArray arrayWithObjects:
				[methodKeyArray objectAtIndex: 0],
				[methodKeyArray objectAtIndex: 1],
				[NSNumber numberWithInt: [keyMethod serial]],
				[keyMethod name],
				nil]; /*Don't forget last nil*/
		[saveKeyArray addObject: tmpArray];
	}
	
	[[NSUserDefaults standardUserDefaults]
		setObject: saveKeyArray forKey: @"Key Bind"];
}

- (void) makeMethodArray
{
int count = 0;

	methodArray = [[NSMutableArray alloc] initWithCapacity: 1];
	[self addMethod:@selector(nextPage)
			name: @"Next Page" serial: count++];
	[self addMethod:@selector(prevPage)
			name: @"Previous Page" serial: count++];
	[self addMethod:@selector(nextLine)
			name: @"Next Line" serial: count++];
	[self addMethod:@selector(prevLine)
			name: @"Previous Line" serial: count++];
	[self addMethod:@selector(nextMail)
			name: @"Next Mail" serial: count++];
	[self addMethod:@selector(prevMail)
			name: @"Previous Mail" serial: count++];
	[self addMethod:@selector(nextPageMail)
			name: @"Next Page Mail" serial: count++];
	[self addMethod:@selector(prevPageMail)
			name: @"Previous Page Mail" serial: count++];
	[self addMethod:@selector(unreadMail)
			name: @"Unread Mail" serial: count++];
	[selectTable noteNumberOfRowsChanged];
}


- (void) loadKeyMethodDict: (NSArray *) saveKeyArray
{
NSEnumerator *enumerator = [saveKeyArray objectEnumerator];
NSArray *oneArray;
NSString *key;
unsigned int modifier;
int serial;

	while (oneArray = [enumerator nextObject])
	{
		key = [oneArray objectAtIndex: 0];
		modifier = [(NSNumber *)[oneArray objectAtIndex: 1] unsignedIntValue];
		serial = [(NSNumber *)[oneArray objectAtIndex: 2] intValue];
		[self keyMethodAddKey: key
				modifierFlags: modifier
				atIndex: serial];
	}
}


- (NSDictionary *) keyMethodDict
{
	return keyMethodDict;
}


- (void) keyMethodInit
{
int index = 0;
	[self keyMethodAddKey: @" " modifierFlags: 0 atIndex: index++];
	[self keyMethodAddKey: @"b" modifierFlags: 0 atIndex: index++];
	[self keyMethodAddKey: @"n" modifierFlags: 0 atIndex: index++];
	[self keyMethodAddKey: @"p" modifierFlags: 0 atIndex: index++];
	[self keyMethodAddKey: @"n" modifierFlags: NSAlternateKeyMask atIndex: index++];
	[self keyMethodAddKey: @"p" modifierFlags: NSAlternateKeyMask atIndex: index++];
}

- (void) keyMethodAddKey: (NSString *)characters
			modifierFlags: (unsigned int)modifierFlags
			atIndex: (int) index
{
MethodKey *methodKey;

	methodKey= [[MethodKey alloc] initWithCharacters: characters
				modifierFlags: modifierFlags];
	[keyMethodDict setObject: [methodArray objectAtIndex: index]
			forKey: [methodKey makeKey]];
	[methodKey autorelease];
}

- (void) addMethod: (SEL) newMethod name: (NSString *) newName
			serial: (int) newSerial
{
KeyMethod *keyMethod = [[KeyMethod alloc]
		initWithMethod: newMethod name: newName serial: newSerial];

	[methodArray addObject: keyMethod];
	[keyMethod autorelease];
}


- (void) setBrowserTextView: newBrowserTextView
{
	browserTextView = newBrowserTextView;
}

- (BOOL) handleKeyDown: (NSEvent *) event
{
MethodKey *methodKey;
KeyMethod *keyMethod;

	methodKey = [[MethodKey alloc] initWithEvent: event];
	keyMethod = (KeyMethod *)[keyMethodDict objectForKey: [methodKey makeKey]];
	if (keyMethod)
	{
		[self performSelector: [keyMethod method]];
		return YES;
	}
	[methodKey autorelease];
	return NO;
}

- (void) nextPage
{
    if ([[scrollView verticalScroller] floatValue] < 0.99999)
        [browserTextView pageDown: self];
    else
        [self nextMail];
}


- (void) prevPage
{
	[browserTextView pageUp: self];
}


- (void) nextLine
{
	[browserTextView moveDown: self];
}


- (void) prevLine
{
	[browserTextView moveUp: self];
}


- (void) nextMail
{
	[mailTitles readNextMail];
}


- (void) prevMail
{
int index = [mailTable selectedRow];

	[mailTable deselectRow: index];
	index--;
	if (index < 0)
		index = 0;
	if ([[mailTitles mailTitles] count])
	{
		[mailTable selectRow: index byExtendingSelection: NO];
		//[table scrollRowToVisible: [mailTitles count]-1]; /*put the selection at top*/
		[mailTable scrollRowToVisible: index];
	}
}


- (void) nextPageMail
{
int index = [mailTable selectedRow];
NSRange range;

	[mailTable deselectRow: index];
	range = [mailTable rowsInRect: [[mailTable superview] visibleRect]];
        index = range.location + range.length - 1;
	if (index >= [[mailTitles mailTitles] count])
		index = [[mailTitles mailTitles] count]-1;
	if ([[mailTitles mailTitles] count])
	{
		[mailTable selectRow: index byExtendingSelection: NO];
		[mailTable scrollRowToVisible: [[mailTitles mailTitles] count]-1]; /*put the selection at top*/
		[mailTable scrollRowToVisible: index];
	}
}


- (void) prevPageMail
{
int index = [mailTable selectedRow];
NSRange range;

	[mailTable deselectRow: index];
	range = [mailTable rowsInRect: [[mailTable superview] visibleRect]];
        index = range.location - range.length + 1;
	if (index < 0)
		index = 0;
	if ([[mailTitles mailTitles] count])
	{
		[mailTable selectRow: index byExtendingSelection: NO];
		[mailTable scrollRowToVisible: 0]; /*put the selection at bottom*/
		[mailTable scrollRowToVisible: index];
	}
}


- (void) unreadMail
{
	[mailTitles readUnreadMail: self];
}


- (IBAction)addKey: (NSMenuItem *) sender
{
	[NSApp stopModal];
	[NSApp runModalForWindow: selectCommandWindow];
}


- (IBAction)enterKeyForCommand: (NSTableView *) sender
{
	[NSApp stopModal];
	currentMethod = (KeyMethod *)[methodArray
			objectAtIndex: [sender selectedRow]];
	[keyWindowMessage setStringValue: [NSString stringWithFormat:
			@"%@ \"%@\"%@", Localize(@"Enter Key for"), [currentMethod name], Localize(@".")]];
    [selectCommandWindow close];
    [menuController allFront: nil];
	[NSApp runModalForWindow: keyWindow];
}


- (IBAction) selectCommandCancel: (NSButton *) sender
{
	[NSApp stopModal];
	[selectCommandWindow close];
	[NSApp runModalForWindow: preferenceWindow];
}

- (IBAction) keyWindowCancel: (NSButton *) sender
{
	[NSApp stopModal];
	[keyWindow close];
	[NSApp runModalForWindow: preferenceWindow];
}


- (int) numberOfRowsInTableView: (NSTableView *)tableView
{
	if (tableView == methodTable)
		return [methodArray count];
	else
		return [keyArray count];
}


- (id) tableView: (NSTableView *)tableView
		objectValueForTableColumn: (NSTableColumn *)tableColumn
		row: (int)row
{
NSString *identifier;
KeyMethod *keyMethod;
MethodKey *methodKey;
NSArray *methodKeyArray;
NSArray *prevMethodKeyArray;

	if (tableView == methodTable)
		return [[methodArray objectAtIndex: row] name];
	else
	{
		identifier = [tableColumn identifier];
		methodKeyArray = (NSArray *)[keyArray objectAtIndex: row];
		if ([identifier isEqualToString: @"commandString"])
		{
			keyMethod = [keyMethodDict objectForKey:
							methodKeyArray];
			if (row <= 0)
				return [keyMethod name];
			prevMethodKeyArray = [keyArray objectAtIndex: row - 1];
			if ([[keyMethod name] isEqualToString:
					[[keyMethodDict objectForKey:
							prevMethodKeyArray]
							name]])
				return @"";
			else
				return [keyMethod name];
		}
		else
		{
			methodKey = [[MethodKey alloc] initWithArray: methodKeyArray];
			[methodKey autorelease];
			return [methodKey valueForKey: identifier];
		}
	}
					
}


- (void) keyDown: (NSEvent *) event
{
MethodKey *methodKey= [[MethodKey alloc] initWithEvent: event];

	[keyMethodDict setObject: currentMethod forKey: [methodKey makeKey]];
	[keyArray autorelease];
	keyArray = [[keyMethodDict allKeys] mutableCopy];
	[methodKey autorelease];
	[keyArray sortUsingFunction: &SortSerial
						context: keyMethodDict];
	[keyTable noteNumberOfRowsChanged];
	if ([keyArray count])
		[keyTable selectRow: 0 byExtendingSelection: NO];
}

- (IBAction) removeSelectionKey: (NSButton *) sender
{
int index = [keyTable selectedRow];

	[keyMethodDict removeObjectForKey: [keyArray objectAtIndex: index]];
	[keyArray removeObjectAtIndex: index];
	[keyTable noteNumberOfRowsChanged];
}


- (IBAction) removeAllKey: (NSButton *) sender
{
	[keyMethodDict removeAllObjects];
	[keyArray removeAllObjects];
	[keyTable noteNumberOfRowsChanged];
}
@end


NSComparisonResult SortSerial(NSArray *methodKeyArray1,
					NSArray *methodKeyArray2, void *context)
{
int serial1, serial2;
NSDictionary *keyMethodDict = (NSDictionary *) context;

	serial1 = [[keyMethodDict objectForKey: methodKeyArray1] serial];
	serial2 = [[keyMethodDict objectForKey: methodKeyArray2] serial];
	return (serial1 > serial2) - (serial1 < serial2);
}
